from django.apps import AppConfig

__all__ = [
    "DjangoXFormulaConfig",
]


class DjangoXFormulaConfig(AppConfig):

    label = "django_xformula"

    name = "django_xformula"

    verbose_name = "XFormula"
