from django_xformula.errors.forbidden_attribute import ForbiddenAttribute
from django_xformula.errors.forbidden_call import ForbiddenCall

__all__ = [
    "ForbiddenAttribute",
    "ForbiddenCall",
]
